from __future__ import annotations

from tach.constants import TOOL_NAME

template = """#!/bin/sh
# Pre-commit script that validates dependencies locally
set -e

{command}"""


def build_pre_commit_hook_content() -> str:
    return template.format(command=f"{TOOL_NAME} check")



from __future__ import annotations

import re
import sys
from functools import lru_cache
from typing import Any

KNOWN_MODULE_SPECIAL_CASES = {
    "__future__",
    "typing_extensions",
}


def is_stdlib_module(module: str) -> bool:
    if module in KNOWN_MODULE_SPECIAL_CASES:
        return True

    if sys.version_info >= (3, 10):
        if module in sys.builtin_module_names:
            return True
        if module in sys.stdlib_module_names:
            return True
        return False
    else:
        from stdlib_list import in_stdlib  # type: ignore

        return in_stdlib(module)  # type: ignore


def get_stdlib_modules() -> list[str]:
    if sys.version_info >= (3, 10):
        modules = set(sys.builtin_module_names)
        modules.update(sys.stdlib_module_names)
        modules.update(KNOWN_MODULE_SPECIAL_CASES)
        return sorted(modules)
    else:
        from stdlib_list import stdlib_list  # type: ignore

        modules: set[str] = set(stdlib_list())  # type: ignore
        modules.update(KNOWN_MODULE_SPECIAL_CASES)
        return list(sorted(modules))


def _get_installed_modules(dist: Any) -> list[str]:
    # This method is best-effort, and is only used for Python < 3.10
    module_names: set[str] = set()

    # Method 1: Check top_level.txt
    try:
        top_level = dist.read_text("top_level.txt")
        if top_level:
            module_names.update(
                module_name.strip()
                for module_name in top_level.splitlines()
                if module_name.strip()
            )
            return list(module_names)
    except Exception:
        pass

    # Method 2: Parse RECORD file
    try:
        record = dist.read_text("RECORD")
        if record:
            for line in record.splitlines():
                base_name = line.split(",")[0].split("/")[0]
                if (
                    "/" in line
                    and not base_name.startswith("_")
                    and not base_name.endswith("dist-info")
                    and "." not in base_name
                ):
                    module_names.add(base_name)
            if module_names:
                return list(module_names)
    except Exception:
        pass

    # Method 3: Check entry points
    for ep in dist.entry_points:
        if ":" in ep.value:
            entry_point = ep.value.split(":")[0]
        else:
            entry_point = ep.value
        module_names.add(entry_point.split(".")[0])
    return list(module_names)


@lru_cache(maxsize=None)
def get_module_mappings() -> dict[str, list[str]]:
    if sys.version_info >= (3, 10):
        from importlib.metadata import packages_distributions

        return packages_distributions()  # type: ignore
    else:
        if sys.version_info >= (3, 8):  # noqa: UP036
            from importlib.metadata import distributions
        else:
            from importlib_metadata import distributions  # type: ignore

        result: dict[str, list[str]] = {}
        for dist in distributions():
            modules = _get_installed_modules(dist)
            name = dist.metadata["Name"]
            for module in modules:
                if module not in result:
                    result[module] = []
                result[module].append(name)
        return result


PYPI_PACKAGE_REGEX = re.compile(r"[-_.]+")


def get_package_name(import_module_path: str) -> str:
    top_level_name = import_module_path.split(".")[0]
    module_mappings = get_module_mappings()
    # Ignoring the case of multiple packages providing this module,
    # using the first one in the mapping
    return module_mappings.get(top_level_name, [top_level_name])[0]


def normalize_package_name(import_module_path: str) -> str:
    return PYPI_PACKAGE_REGEX.sub("-", get_package_name(import_module_path)).lower()


__all__ = [
    "is_stdlib_module",
    "get_module_mappings",
    "get_package_name",
    "normalize_package_name",
]



from __future__ import annotations

from tach.parsing.config import parse_project_config
from tach.show import generate_show_report


# right now this is just a smoke test
# this example directory has Python files outside source roots, which has previously caused bugs
def test_many_features_example_dir(example_dir, capfd):
    project_root = example_dir / "many_features"
    project_config = parse_project_config(root=project_root)
    assert project_config is not None

    report = generate_show_report(
        project_root=project_root, project_config=project_config, included_paths=[]
    )
    assert report is not None



from __future__ import annotations

import json
import os
import subprocess
import sys
import time
from http.client import HTTPSConnection
from pathlib import Path
from typing import Any
from urllib import error, parse, request

LOGGING_URL = "https://vmilasesnyvpalekembc.supabase.co"
PUBLIC_ANON_CLIENT_KEY = (
    "eyJhbGciOiJIUzI1NiIsInR5cCI6IkpXVCJ9.eyJpc3MiOiJzdXBhYmFzZSIsInJlZiI6InZtaWxhc2Vzbnl2cGFsZWtlbWJjIiwicm9"
    "sZSI6ImFub24iLCJpYXQiOjE3MTY0OTEwOTMsImV4cCI6MjAzMjA2NzA5M30.ndk9sUAmMJ5oNenDmLw35uT0s_d6c56Hk_PL5BucrOc"
)


def update_latest_version(project_root: Path) -> None:
    url = "https://pypi.org/pypi/tach/json"
    try:
        with request.urlopen(url, timeout=1) as response:
            if response.status == 200:
                data = response.read().decode()
                json_data = json.loads(data)
                latest_version = json_data["info"]["version"]
            else:
                return
    except (error.URLError, KeyError):
        return
    (project_root / ".tach" / ".latest-version").write_text(latest_version)


def log_request(url: str, data: dict[str, Any]) -> None:
    headers = {
        "Content-Type": "application/json",
        "apikey": PUBLIC_ANON_CLIENT_KEY,
        "authorization": f"Bearer {PUBLIC_ANON_CLIENT_KEY}",
    }
    json_data = json.dumps(data)
    full_url = f"{LOGGING_URL}/{url}"
    conn = None
    try:
        url_parts: parse.ParseResult = parse.urlparse(full_url)
        conn = HTTPSConnection(url_parts.netloc, timeout=1)
        conn.request("POST", full_url, body=json_data, headers=headers)
        conn.getresponse()
    except Exception:  # noqa
        pass
    finally:
        if conn is not None:
            conn.close()


def log_uid(uid: str, is_ci: bool, is_gauge: bool) -> None:
    log_request(
        url="rest/v1/User", data={"id": uid, "is_ci": is_ci, "is_gauge": is_gauge}
    )


def log_record(record_data: dict[str, Any]) -> None:
    log_request(url="rest/v1/LogRecord", data=record_data)


def send_log_entry(
    version: str,
    uid: str | None,
    function: str,
    parameters: dict[str, Any],
    level: str,
    timestamp: float,
    entry: str,
) -> None:
    is_ci = "CI" in os.environ
    is_gauge = "IS_GAUGE" in os.environ
    log_data: dict[str, Any] = {
        "user": str(uid) if uid else None,
        "message": entry,
        "level": level,
        "timestamp": timestamp,
        "function": function,
        "parameters": parameters,
        "version": version,
    }
    if uid:
        log_uid(uid=uid, is_ci=is_ci, is_gauge=is_gauge)
    log_record(log_data)


def process_message(message: dict[str, Any]) -> None:
    version = message["version"]
    uid = message["uid"]
    function = message["call_info"]["function"]
    parameters = message["call_info"]["parameters"]
    level = message["level"]
    timestamp = message["timestamp"]
    entry = message["log_entry"]
    send_log_entry(version, uid, function, parameters, level, timestamp, entry)


def subprocess_worker(file_path: Path, timeout: int = 5) -> None:
    try:
        last_position = 0
        last_message_time = time.time()

        while True:
            try:
                with open(file_path) as f:
                    f.seek(last_position)
                    lines = f.readlines()

                    if lines:
                        last_message_time = time.time()
                        last_position = f.tell()

                        for line in lines:
                            try:
                                message = json.loads(line)
                                process_message(message)
                            except json.JSONDecodeError:
                                continue
                            except Exception:
                                pass

            except Exception:
                pass

            time.sleep(0.1)
            if time.time() - last_message_time > timeout:
                break

    finally:
        if file_path.exists():
            file_path.unlink()


def create_managed_subprocess(project_root: Path, timeout: int = 5) -> Path:
    """
    Launches the worker as a completely separate process using subprocess.Popen.
    Returns the path to the temporary file for message passing.
    """
    tach_dir = project_root / ".tach"
    tach_dir.mkdir(parents=True, exist_ok=True)

    # Cannot use FIFO because it is not supported on Windows
    log_file = tach_dir / "log_pipe.txt"
    if log_file.exists():
        log_file.unlink()
    log_file.touch()

    worker_script = Path(__file__).resolve()
    subprocess.Popen(
        [sys.executable, str(worker_script), "--worker", str(log_file), str(timeout)],
        start_new_session=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )

    return log_file


# This is the entrypoint from subprocess.Popen in create_managed_subprocess
if __name__ == "__main__":
    if len(sys.argv) > 1 and sys.argv[1] == "--worker":
        fifo_path = Path(sys.argv[2])
        timeout = int(sys.argv[3])
        subprocess_worker(fifo_path, timeout)



from __future__ import annotations

import pathlib
from pathlib import Path
from unittest.mock import Mock

import pytest

from tach import cli
from tach.extension import ProjectConfig

_VALID_TACH_TOML = pathlib.Path(__file__).parent / "example" / "valid" / "tach.toml"


@pytest.fixture
def mock_check(mocker) -> Mock:
    mock = Mock(return_value=[])  # default to a return with no errors
    mocker.patch("tach.extension.check", mock)
    return mock


@pytest.fixture
def mock_project_config(mocker) -> ProjectConfig:
    def _mock_project_config(root: str = "") -> ProjectConfig:
        return ProjectConfig()

    mocker.patch("tach.cli.parse_project_config", _mock_project_config)
    return _mock_project_config()


def test_execute_with_config(capfd, mock_check, mock_project_config):
    # Test with a valid path as mocked
    args, _ = cli.parse_arguments(["check"])
    assert args.command == "check"
    with pytest.raises(SystemExit) as sys_exit:
        cli.tach_check(
            project_root=Path(),
            project_config=mock_project_config,
        )
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out


def test_invalid_command(capfd):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with an invalid command
        cli.parse_arguments(["help"])
    captured = capfd.readouterr()
    assert sys_exit.value.code == 2
    assert "invalid choice: 'help" in captured.err


def test_execute_with_valid_exclude(capfd, mock_check, mock_project_config):
    with pytest.raises(SystemExit) as sys_exit:
        # Test with a valid path as mocked
        args, _ = cli.parse_arguments(["check", "--exclude", "valid_dir"])
        mock_project_config.exclude = args.exclude.split(",")
        cli.tach_check(
            project_root=Path(),
            project_config=mock_project_config,
        )
    captured = capfd.readouterr()
    assert sys_exit.value.code == 0
    assert "✅" in captured.out
    assert "All modules validated!" in captured.out


def test_tach_server_with_config(tmp_path, mocker):
    mock_run_server = mocker.patch("tach.extension.run_server", autospec=True)

    cli.main(["server", "--config", str(_VALID_TACH_TOML)])
    # Verify server was run with the custom config.
    mock_run_server.assert_called_once()
    assert "domain_four.py" in mock_run_server.call_args[0][1].exclude

    mock_run_server.reset_mock()

    # Should still work even if it's not named tach.toml
    toml_contents = _VALID_TACH_TOML.read_text()
    custom_config_path = tmp_path.joinpath("custom_config.toml")
    custom_config_path.write_text(toml_contents)
    cli.main(["server", "--config", str(custom_config_path)])

    # Verify server was run with the custom config.
    mock_run_server.assert_called_once()
    assert "domain_four.py" in mock_run_server.call_args[0][1].exclude



from __future__ import annotations

import json
import os
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING

from tach.errors import TachError, TachSetupError

if TYPE_CHECKING:
    from git import Repo


@dataclass
class GitBranchInfo:
    repo: str
    name: str
    commit: str
    owner: str
    email: str


def is_github_actions():
    return os.environ.get("GITHUB_ACTIONS") == "true"


def _get_branch_name(repo: Repo) -> str:
    # GHA uses a detached HEAD / shallow clone in actions/checkout@v4
    if is_github_actions():
        event_name = os.environ.get("GITHUB_EVENT_NAME")
        # Different environment variables are used for PRs and pushes
        if event_name == "pull_request":
            return os.environ["GITHUB_HEAD_REF"]
        return os.environ["GITHUB_REF_NAME"]
    else:
        return repo.active_branch.name


def _get_commit(repo: Repo) -> str:
    # GHA uses a detached HEAD / shallow clone in actions/checkout@v4
    if is_github_actions():
        event_name = os.environ.get("GITHUB_EVENT_NAME")
        event_path = os.environ.get("GITHUB_EVENT_PATH")
        if event_path and event_name == "pull_request":
            with open(event_path) as f:
                event_info = json.load(f)
                # Pull commit from the PR event in GHA
                return event_info["pull_request"]["head"]["sha"]
    return repo.head.commit.hexsha


def _get_owner_and_repo_name(repo_url: str) -> tuple[str, str]:
    """
    Parse a git URL to extract owner and repo names, handling cases where
    either might contain slashes.

    Strategy:
    1. For HTTPS URLs, everything after the domain and before the last segment is the owner
    2. For SSH URLs, everything after the colon and before the last segment is the owner
    3. The last segment (minus .git) is always the repo name

    Args:
        url: Git remote URL (HTTPS or SSH format)

    Returns:
        Tuple of (owner, repo)

    Examples:
        >>> _get_owner_and_repo_name("https://github.com/apache/spark/spark-connector.git")
        ('apache/spark', 'spark-connector')
        >>> _get_owner_and_repo_name("git@github.com:facebook/react-native/docs.git")
        ('facebook/react-native', 'docs')
    """
    # Remove .git suffix and slashes
    url = repo_url.strip("/")

    if url.endswith(".git"):
        url = url[:-4].strip("/")

    if url.startswith("https://"):
        # HTTPS URL format
        # Remove https:// prefix and split by /
        parts = url[8:].split("/")
        # First part is domain (e.g., github.com)
        # Last part is repo name
        repo = parts[-1]
        # Everything between domain and repo is owner
        owner = "/".join(parts[1:-1])
    else:
        # SSH URL format
        # Split at the colon
        parts = url.split(":")
        if len(parts) != 2:
            raise TachError(f"Invalid SSH URL format: {url}")
        path_parts = parts[1].split("/")
        # Last part is repo name
        repo = path_parts[-1]
        # Everything else is owner
        owner = "/".join(path_parts[:-1])

    if not owner or not repo:
        raise TachError(f"Failed to parse owner or repo from URL: {url}")

    return owner, repo


def get_current_branch_info(
    project_root: Path, allow_dirty: bool = False
) -> GitBranchInfo:
    # Local import because git-python takes ~80ms to load
    from git import InvalidGitRepositoryError, NoSuchPathError, Repo

    try:
        repo = Repo(project_root, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        raise TachSetupError("The project does not appear to be a git repository!")

    # Fail if the branch is not clean
    if not allow_dirty and repo.is_dirty():
        raise TachError(
            "The current branch is not clean, please commit your changes before running this command!"
        )

    try:
        url = repo.remotes.origin.url
        owner_name, repo_name = _get_owner_and_repo_name(url)
        config_reader = repo.config_reader()
        email = str(config_reader.get_value("user", "email", default=""))
        branch = _get_branch_name(repo)
        commit = _get_commit(repo)

    except Exception as e:
        raise TachError(f"Failed to determine current branch information!\nError: {e}")

    return GitBranchInfo(
        repo=repo_name,
        owner=owner_name,
        name=branch,
        commit=commit,
        email=email,
    )


def get_changed_files(
    project_root: Path, head: str = "", base: str = "main"
) -> list[Path]:
    # Local import because git-python takes ~80ms to load
    from git import GitCommandError, InvalidGitRepositoryError, NoSuchPathError, Repo

    try:
        repo = Repo(project_root, search_parent_directories=True)
    except (InvalidGitRepositoryError, NoSuchPathError):
        raise TachSetupError(
            "The project does not appear to be a git repository, cannot determine changed files!"
        )

    try:
        if head:
            diff: str = repo.git.diff("--name-status", head, base)
        else:
            # If head is not provided, we can diff against 'base' from the current filesystem
            diff: str = repo.git.diff("--name-status", base)
    except GitCommandError:
        head_display = f"'{head}'" if head else "current filesystem"
        raise TachError(f"Failed to check diff between '{base}' and {head_display}!")

    diff_lines = diff.splitlines()
    changed_files: set[str] = set()

    for line in diff_lines:
        _, *files = line.split("\t")
        changed_files.update(files)

    if not head:
        # If we are using the current filesystem, there may be relevant changes in untracked files
        untracked_files: str = repo.git.ls_files("--others", "--exclude-standard")
        changed_files.update(untracked_files.splitlines())

    # return list of unique Paths
    git_root: str = repo.git.rev_parse("--show-toplevel")
    return [(Path(git_root) / filepath).resolve() for filepath in changed_files]


__all__ = ["get_changed_files", "get_current_branch_info"]



from __future__ import annotations

import uuid
from pathlib import Path

from tach import __version__
from tach.filesystem import find_project_config_root


def resolve_dot_tach() -> Path | None:
    project_path = find_project_config_root()
    if project_path is None:
        return

    def _create(path: Path, is_file: bool = False, file_content: str = "") -> None:
        if not path.exists():
            if is_file:
                path.write_text(file_content.strip())
            else:
                path.mkdir()

    # Create .tach
    tach_path = project_path / ".tach"
    _create(tach_path)
    # Create info
    info_path = tach_path / "tach.info"
    _create(info_path, is_file=True, file_content=str(uuid.uuid4()))
    # Create .gitignore
    gitignore_content = """
# This folder is for tach. Do not edit.

# gitignore all content, including this .gitignore
*
    """
    gitignore_path = tach_path / ".gitignore"
    _create(gitignore_path, is_file=True, file_content=gitignore_content)
    # Create version
    version_path = tach_path / ".latest-version"
    _create(version_path, is_file=True, file_content=__version__)
    return Path(tach_path)



from __future__ import annotations


class BCOLORS:
    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"


__all__ = ["BCOLORS"]



from __future__ import annotations

from typing import TYPE_CHECKING

from rich.panel import Panel
from rich.prompt import Confirm

from tach import errors
from tach import filesystem as fs
from tach.console import console
from tach.constants import CONFIG_FILE_NAME, TOOL_NAME
from tach.extension import ProjectConfig, parse_project_config, sync_project
from tach.mod import mod_edit_interactive
from tach.show import upload_show_report

if TYPE_CHECKING:
    from pathlib import Path


def mark_modules(project_root: Path, project_config: ProjectConfig) -> ProjectConfig:
    saved, _ = mod_edit_interactive(
        project_root=project_root,
        project_config=project_config,
        exclude_paths=project_config.exclude,
    )
    if not saved:
        raise errors.TachInitCancelledError()

    project_config_path = fs.build_project_config_path(project_root)
    project_config, _ = parse_project_config(project_config_path)
    return project_config


def sync_modules(project_root: Path, project_config: ProjectConfig) -> ProjectConfig:
    sync_project(project_root, project_config)

    project_config_path = fs.build_project_config_path(project_root)
    project_config, _ = parse_project_config(project_config_path)
    return project_config


def prompt_to_re_select_modules() -> bool:
    console.print(
        Panel(
            "No dependencies found between the selected modules.\n"
            "This might mean that your source root is not set correctly,"
            " or you did not select modules which depend on each other.",
            style="yellow",
        )
    )
    return Confirm.ask(
        "[cyan]Would you like to re-select modules?[/]\n",
        default=True,
        show_default=False,
    )


def prompt_to_show_project() -> bool:
    console.print(
        Panel(
            "Would you like to visualize your dependency graph?\n"
            f"This will upload your module configuration in [cyan]'{CONFIG_FILE_NAME}.toml'[/] to Gauge ([blue underline]https://app.gauge.sh[/])",
            style="yellow",
        )
    )
    return Confirm.ask("", default=False, show_default=False)


def show_project(project_config: ProjectConfig, project_root: Path):
    if prompt_to_show_project():
        show_url = upload_show_report(project_root, project_config, included_paths=[])
        if show_url:
            console.print(
                "\n[cyan]View your dependency graph here:[/]\n"
                f"[blue underline]{show_url}[/]"
            )
        else:
            console.print(
                "[red]Failed to generate show URL. Please try again later.[/]"
            )


def setup_modules(project_root: Path, project_config: ProjectConfig) -> ProjectConfig:
    project_config = mark_modules(project_root, project_config)
    project_config = sync_modules(project_root, project_config)

    while project_config.has_no_dependencies():
        wants_to_re_select = prompt_to_re_select_modules()
        if wants_to_re_select:
            project_config = mark_modules(project_root, project_config)
            project_config = sync_modules(project_root, project_config)
        else:
            console.print("[cyan]Continuing with selected modules.[/]")
            break
    return project_config


def get_all_existing_config_files(project_root: Path) -> list[Path]:
    current_config_path = fs.get_project_config_path(project_root)

    def exclude_self(path: Path) -> bool:
        # This is a quick proxy to exclude Tach's own config files in the venv
        return "/tach/" not in str(path)

    return list(
        filter(
            None,
            (
                current_config_path,
                *filter(exclude_self, project_root.rglob("tach.domain.toml")),
            ),
        ),
    )


def init_project(project_root: Path, force: bool = False):
    config_files = get_all_existing_config_files(project_root)
    if config_files:
        if not force:
            raise errors.TachError(
                f"[yellow]Project already initialized. Use [cyan]`{TOOL_NAME} init --force`[/] to reinitialize.[/]"
            )

        console.print(
            Panel(
                "Project already initialized. Would you like to reinitialize?\n"
                "This will overwrite the current configuration.",
                style="yellow",
            )
        )

        if Confirm.ask("", default=False, show_default=False):
            try:
                for config_file in config_files:
                    config_file.unlink()
            except OSError:
                raise errors.TachError("[red]Failed to remove configuration file.[/]")
        else:
            raise errors.TachError(
                "[red]Refusing to overwrite existing project configuration.[/]"
            )

    console.print(
        Panel(
            "Welcome to Tach! Let's get started by selecting the modules you want to include in your project.\n"
            "We will use [cyan]'tach mod'[/] to interactively mark which files or folders should be tracked. You can learn more at [blue underline]https://docs.gauge.sh/usage/faq[/]",
            style="yellow",
        )
    )
    console.print("\nPress [cyan]'Enter'[/] to get started.", style="cyan")
    console.input()

    project_config = ProjectConfig()

    try:
        project_config = setup_modules(project_root, project_config)
    except errors.TachInitCancelledError:
        console.print("[yellow]Initialization cancelled.[/]")
        return

    show_project(project_config, project_root)

    console.print(
        "\n[green]Tach is now configured for this project!"
        " Run [cyan]'tach check'[/] to validate your configuration.[/]\n\n"
        "[green]Full documentation is available at [blue underline]https://docs.gauge.sh[/]"
    )



# tach-ignore
from module2.service import Module2Service

# tach-ignore(ok)
from module1 import (
    a,  # tach-ignore(ok)
    b,
    c,
    d,
    e,
    f,
    g,  # tach-ignore(ok)
    h,
    i,
    j,
)


from pathlib import Path

from setuptools import setup


# For this trick, I thank:
# https://stackoverflow.com/a/78056725

root: Path = Path(__file__).parent
pack_d_path: str = (root.parent / "pack-d").as_uri()
pack_e_path: str = (root.parent / "pack-e").as_uri()

if __name__ == "__main__":
    setup(
        install_requires=[
            f"myorg-pack-d @ {pack_d_path}",
            f"myorg-pack-e @ {pack_e_path}",
        ],
    )
